#include <csignal>
#include <cstring>
#include <sstream>
#include <string>
#include <thread>

#include "common/file_parser/helper_h264_parser.h"
#include "common/helper.h"
#include "common/opt_parser.h"
#include "common/sample_common.h"
#include "common/sample_connection_observer.h"
#include "common/sample_event.h"
#include "common/sample_local_user_observer.h"
#include "common/write_csvfile.h"
#include "common/log.h"

#define DEFAULT_CONNECT_TIMEOUT_MS (3000)
#define DEFAULT_FRAME_RATE (30)
#define DEFAULT_VIDEO_FILE "test_data/test_multi_slice.h264"
#define DEFAULT_RECEIVE_DATASTREAM_INFO "start"
#define DEFAULT_STATISTICS_COUNT (100)
#define DEFAULT_DATASTREAM_WAIT_TIME -1

static uint64_t receive_datastream_T1 = 0;
static uint64_t send_one_h264_T2 = 0;
static uint64_t round_count = 0;
static uint64_t sumSpendTime = 0;
static int statistics_count_per_time = 0;
static SampleEvent receive_ready_;
static std::string required_receive_datastream =
    DEFAULT_RECEIVE_DATASTREAM_INFO;
static uint64_t FrameCount = 0;
static bool error_flag = false;

static std::unique_ptr<WriteCSVFileHandle> csv_file_handle = nullptr;

struct SampleOptions {
  std::string appId;
  std::string channelId;
  std::string userId;

  std::string videoFile = DEFAULT_VIDEO_FILE;
  std::string datastream_info = DEFAULT_RECEIVE_DATASTREAM_INFO;
  int datastream_wait_time = DEFAULT_DATASTREAM_WAIT_TIME;
  struct {
    int frameRate = DEFAULT_FRAME_RATE;
  } video;
  int rsfec_min_level = 0;
  int statistics_count = DEFAULT_STATISTICS_COUNT;
  std::string vos_ip;
};

static int waitUntilReceiveMessage(int waitMs) {
  return receive_ready_.Wait(waitMs);
}

static void sendOneH264Frame(
    int frameRate, std::unique_ptr<HelperH264Frame> h264Frame,
    agora::agora_refptr<agora::rtc::IVideoEncodedImageSender>
        videoH264FrameSender) {
  agora::rtc::EncodedVideoFrameInfo videoEncodedFrameInfo;
  videoEncodedFrameInfo.rotation = agora::rtc::VIDEO_ORIENTATION_0;
  videoEncodedFrameInfo.codecType = agora::rtc::VIDEO_CODEC_H264;
  videoEncodedFrameInfo.framesPerSecond = frameRate;
  videoEncodedFrameInfo.frameType =
      (h264Frame.get()->isKeyFrame
           ? agora::rtc::VIDEO_FRAME_TYPE::VIDEO_FRAME_TYPE_KEY_FRAME
           : agora::rtc::VIDEO_FRAME_TYPE::VIDEO_FRAME_TYPE_DELTA_FRAME);

  FrameCount++;

  printf(
      "[ %s ]--> Now  Send %luth Frame H264 Data, length=%d,  Send "
      "frameRate= %d\n ",
      getCurrentSystemTimeChrono().c_str(), FrameCount,
      h264Frame.get()->bufferLen, frameRate);
  send_one_h264_T2 = 0;
  send_one_h264_T2 = now_ms_t();

  videoH264FrameSender->sendEncodedVideoImage(
      reinterpret_cast<uint8_t*>(h264Frame.get()->buffer.get()),
      h264Frame.get()->bufferLen, videoEncodedFrameInfo);
  if (!error_flag) {
    int spendTime = send_one_h264_T2 - receive_datastream_T1;
    round_count++;
    printf("recv datastream then send one h264 round_count:%lu, spendTime:%d\n",
           round_count, spendTime);

    delayTimeBaseInfo delay_info;
    delay_info.round_count_ = round_count;
    delay_info.spendTime_ = spendTime;
    delay_info.data_length = h264Frame.get()->bufferLen;
    csv_file_handle->write_csvfile(delay_info);

    sumSpendTime += spendTime;
    if (round_count % statistics_count_per_time == 0) {
      double averageSpendTime =
          sumSpendTime / (double)statistics_count_per_time;
      sumSpendTime = 0;
      printf(
          "the average spendTime from receive datastream to send h264 per %d "
          "times is %f \n",
          statistics_count_per_time, averageSpendTime);
    }
  }
}

static void SampleSendVideoH264Task(
    const SampleOptions& options,
    agora::agora_refptr<agora::rtc::IVideoEncodedImageSender>
        videoH264FrameSender,
    bool& exitFlag) {
  std::unique_ptr<HelperH264FileParser> h264FileParser(
      new HelperH264FileParser(options.videoFile.c_str()));
  h264FileParser->initialize();

  // Calculate send interval based on frame rate. H264 frames are sent at this
  // interval
  PacerInfo pacer = {0, 1000 / options.video.frameRate,
                     std::chrono::steady_clock::now()};

  while (!exitFlag) {
    error_flag = false;
    int iret = waitUntilReceiveMessage(options.datastream_wait_time);
    if (iret < 0 || exitFlag) {
      continue;
    }
    pacer = {0, 1000 / options.video.frameRate,
             std::chrono::steady_clock::now()};
    std::unique_ptr<HelperH264Frame> h264Frame = nullptr;
    while (h264Frame == nullptr) {
      h264Frame = h264FileParser->getH264Frame();
    }
    sendOneH264Frame(options.video.frameRate, std::move(h264Frame),
                     videoH264FrameSender);
    waitBeforeNextSend(pacer);  // sleep for a while before sending next frame
  };
}

class SampleConnectionObserverHandle : public SampleConnectionObserver {
 public:
  SampleConnectionObserverHandle();
  virtual ~SampleConnectionObserverHandle();

 public:
  // inherit from agora::rtc::IRtcConnectionObserver

  void onStreamMessageError(agora::user_id_t userId, int streamId, int code,
                            int missed, int cached) override;
};

SampleConnectionObserverHandle::SampleConnectionObserverHandle() {}

SampleConnectionObserverHandle::~SampleConnectionObserverHandle() {}


class SampleLocalUserObserverHandle : public SampleLocalUserObserver {
public:
  SampleLocalUserObserverHandle(agora::rtc::ILocalUser* local_user);
  virtual ~SampleLocalUserObserverHandle();
public:
  // inherit from agora::rtc::ILocalUserObserver
  void onStreamMessage(agora::user_id_t userId, int streamId, const char *data,
                        size_t length) override;
};

SampleLocalUserObserverHandle::SampleLocalUserObserverHandle(agora::rtc::ILocalUser* local_user)
  :SampleLocalUserObserver(local_user){}
SampleLocalUserObserverHandle::~SampleLocalUserObserverHandle(){}


void SampleLocalUserObserverHandle::onStreamMessage(agora::user_id_t userId,
                                                     int streamId,
                                                     const char* data,
                                                     size_t length) {
  unsigned int uid = static_cast<unsigned int>(atoll(userId));
  auto key = std::make_pair(uid, streamId);

  if (data == nullptr || length == 0) {
    printf("*** Data stream received fail, data = %p, length = %d\n", data,
           static_cast<int>(length));
    error_flag = true;
    return;
  }

  int req_data_len = required_receive_datastream.size();
  if (strncmp(data, required_receive_datastream.c_str(), req_data_len) != 0) {
    printf("receive error content !\n");
    error_flag = true;
    return;
  }
  printf(
      "[ %s ]--> Now  Receive a Data Stream Message,  Data stream received:"
      "userId = %s, streamId = %d, length = %d\n",
      getCurrentSystemTimeChrono().c_str(), userId, streamId,
      static_cast<int>(length));
  receive_datastream_T1 = 0;
  receive_datastream_T1 = now_ms_t();
  receive_ready_.Set();
}

void SampleConnectionObserverHandle::onStreamMessageError(
    agora::user_id_t userId, int streamId, int code, int missed, int cached) {
  printf(
      "*** Receive StreamMessage error, userId = %s, streamId = %d, code = "
      "%d, missed = %d, cached = %d\n",
      userId, streamId, code, missed, cached);
  error_flag = true;
}

static bool exitFlag = false;
static void SignalHandler(int sigNo) {
  exitFlag = true;
  receive_ready_.Set();
}

int main(int argc, char* argv[]) {
  SampleOptions options;
  opt_parser optParser;

  optParser.add_long_opt("token", &options.appId,
                         "The token for authentication / must");
  optParser.add_long_opt("channelId", &options.channelId, "Channel Id / must");
  optParser.add_long_opt("userId", &options.userId, "User Id / default is 0");
  optParser.add_long_opt("videoFile", &options.videoFile,
                         "The video file1 in H264 format to be sent");
  optParser.add_long_opt("fps", &options.video.frameRate,
                         "Target frame rate for sending the video stream");
  optParser.add_long_opt("datastream_info", &options.datastream_info,
                         "The content you want to receive datastream");
  optParser.add_long_opt("rsfec_min_level", &options.rsfec_min_level,
                         "Set the fec level[0 | 1 | 2 | 3 ]");
  optParser.add_long_opt("statistics_count", &options.statistics_count,
                         "statistics count per time");
  optParser.add_long_opt("vos_ip", &options.vos_ip,
                         "set the vos ip:port address ");

  if ((argc <= 1) || !optParser.parse_opts(argc, argv)) {
    std::ostringstream strStream;
    optParser.print_usage(argv[0], strStream);
    std::cout << strStream.str() << std::endl;
    return -1;
  }

  if (options.appId.empty()) {
    AG_LOG(ERROR, "Must provide appId!");
    return -1;
  }

  if (options.channelId.empty()) {
    AG_LOG(ERROR, "Must provide channelId!");
    return -1;
  }
  statistics_count_per_time = options.statistics_count;
  std::signal(SIGQUIT, SignalHandler);
  std::signal(SIGABRT, SignalHandler);
  std::signal(SIGINT, SignalHandler);

  if (options.datastream_info.size() > 0) {
    required_receive_datastream = options.datastream_info;
  }

  // Create Agora service
  agora::base::IAgoraService* service = createAndInitAgoraService(false, true, true);
  if (!service) {
    AG_LOG(ERROR, "Failed to creating Agora service!");
    return -1;
  }
  // create connection
  agora::rtc::RtcConnectionConfiguration ccfg;
  ccfg.clientRoleType = agora::rtc::CLIENT_ROLE_BROADCASTER;
  ccfg.channelProfile = agora::CHANNEL_PROFILE_CLOUD_GAMING;
  ccfg.autoSubscribeAudio = false;
  ccfg.autoSubscribeVideo = false;
  ccfg.audioRecvMediaPacket = true;
  ccfg.videoRecvMediaPacket = true;
  agora::agora_refptr<agora::rtc::IRtcConnection> connection =
      service->createRtcConnection(ccfg);
  if (!connection) {
    AG_LOG(ERROR, "Failed to creating Agora connection!");
    return -1;
  }

  // set vos server ip address
  agora::base::IAgoraParameter* agoraParameter =
      connection->getAgoraParameter();
  if (options.vos_ip.size() > 0) {
    std::string vos_ip_info = "\"" + options.vos_ip + "\"";
    std::string vos_list = "{\"rtc.vos_list\":[" + vos_ip_info + "]}";
    agoraParameter->setParameters(vos_list.c_str());
  }
  auto ConnObserver = std::make_shared<SampleConnectionObserverHandle>();
  connection->registerObserver(ConnObserver.get());
  // Create local user observer to monitor intra frame request
  auto localUserObserver = std::make_shared<SampleLocalUserObserverHandle>(connection->getLocalUser());
  // Connect to Agora channel
  if (connection->connect(options.appId.c_str(), options.channelId.c_str(),
                          options.userId.c_str())) {
    AG_LOG(ERROR, "Failed to connect to Agora channel!");
    return -1;
  }

  printf("start send h264 and pcm data!\n");
  // Create media node factory
  agora::agora_refptr<agora::rtc::IMediaNodeFactory> factory =
      service->createMediaNodeFactory();
  if (!factory) {
    AG_LOG(ERROR, "Failed to create media node factory!");
  }

  // Create video frame sender
  agora::agora_refptr<agora::rtc::IVideoEncodedImageSender> videoFrameSender =
      factory->createVideoEncodedImageSender();
  if (!videoFrameSender) {
    AG_LOG(ERROR, "Failed to create video frame sender!");
    return -1;
  }

  agora::base::SenderOptions option;
  option.ccMode = agora::base::CC_ENABLED;
  // Create video track
  agora::agora_refptr<agora::rtc::ILocalVideoTrack> customVideoTrack =
      service->createCustomVideoTrack(videoFrameSender, option);
  if (!customVideoTrack) {
    AG_LOG(ERROR, "Failed to create video track!");
    return -1;
  }
  ConnObserver->waitUntilConnected(DEFAULT_CONNECT_TIMEOUT_MS);
  // Publish video track
  connection->getLocalUser()->publishVideo(customVideoTrack);

  std::unique_ptr<WriteCSVFileHandle> csv_handle_(new WriteCSVFileHandle(
      "./out_file/receive_datastream_send_h264_delay.csv"));
  csv_file_handle = std::move(csv_handle_);
  csv_file_handle->open();

  // Start sending media data
  AG_LOG(INFO, "Start sending video data ...");
  std::thread sendVideoThread(SampleSendVideoH264Task, options,
                              videoFrameSender, std::ref(exitFlag));

  sendVideoThread.join();

  // Unpublish video track
  connection->getLocalUser()->unpublishVideo(customVideoTrack);
  // Unregister connection observer
  connection->unregisterObserver(ConnObserver.get());

  // Disconnect from Agora channel
  if (connection->disconnect()) {
    AG_LOG(ERROR, "Failed to disconnect from Agora channel!");
    return -1;
  }

  AG_LOG(INFO, "Disconnected from Agora channel successfully");
  ConnObserver.reset();
  localUserObserver.reset();
  videoFrameSender = nullptr;
  customVideoTrack = nullptr;
  factory = nullptr;
  connection = nullptr;
  csv_file_handle->close();
  csv_file_handle.reset();
  return 0;
}
